from random import shuffle
from model.requirements import RequirementSameTeam, RequirementDifferentTeam


class ImpossibleDistributionError(Exception):
    pass

class DistributionMaker:
    PREFER_SAME_TEAMS = False
    PREFER_DIFFERENT_TEAMS = True

    def __init__(self, group, requirements, team_count=None, min_team_size=None, max_team_size=None, prefer=None):
        self.requirements_same = [
            requirement for requirement in requirements if isinstance(requirement, RequirementSameTeam)
        ]
        self.requirements_different = [
            requirement for requirement in requirements if isinstance(requirement, RequirementDifferentTeam)
        ]
        self.team_count = team_count
        self.min_team_size = min_team_size if min_team_size is not None else 0
        self.max_team_size = max_team_size if max_team_size is not None else group.member_count
        self.prefer = prefer
        self.members = group.members
        self.unassigned = [{member} for member in group.members]
        self.member_count = group.member_count

        self.initMemberBlackList(self.requirements_different)

    def initMemberBlackList(self, requirements):
        for member in self.members:
            member.blacklist_ = set()
        for requirement in requirements:
            requirement.member.blacklist_ |= set(requirement.target_members)
            for member in requirement.target_members:
                member.blacklist_ |= {requirement.member} | \
                                     set(person for person in requirement.target_members if person != member)

    def distribute(self):
        """
        create distribution
        :return: map of teams (consisting of members) - {"team1": {Member1, Member2,...}, "team2": {Member, ...}}
        :raise ImpossibleDistributionError() if the distribution with required parameters doesn't exist
        """
        if not self.checkTeamSizeLimits():
            raise ImpossibleDistributionError()

        self._makeTeamsFromRequirements()

        for teams in self._solve(self.unassigned):
            if self.team_count is None or len(teams) == self.team_count:
                for team in teams:
                    if not self.min_team_size <= len(team) <= self.max_team_size:
                        break
                else:
                    self.printTeams(teams)
                    res = {}
                    shuffle(teams)
                    for team_nr, team in enumerate(teams):
                        res["Tým " + str(team_nr)] = list(teams[team_nr])
                    return res
        raise ImpossibleDistributionError()

    def checkTeamSizeLimits(self):
        """
        check whether the team size limits (max_team_size, min_team_size, team_count) are satisfiable
        :return: True if satisfiable, False if not
        """
        if self.max_team_size < self.min_team_size:
            return False

        if self.team_count is not None:
            avg_team_size = self.member_count / self.team_count
            if self.min_team_size > avg_team_size:
                return False
            if self.max_team_size < avg_team_size:
                return False
        return True

    def checkTeamValid(self, team):
        """
        checks if the team is not longer than max_team_size and doesn't violate any requirement
        Warning: the team can still be shorter than min_team_size!
        :param team: 
        :return: True if the team is valid, False otherwise
        """
        if len(team) > self.max_team_size:
            return False
        for member in team:
            for member_blacklist in member.blacklist_:
                if member_blacklist in team:
                    return False
        return True

    def _makeTeamsFromRequirements(self):
        """
        make teams (sets) from members in self.unassigned based on requirements of 'must be in the same team' 
        """
        teams = []
        for requirement in self.requirements_same:
            team = set(requirement.target_members + [requirement.member])
            team_nr = None
            for member in team:
                # remove member from unassigned
                try:
                    self.unassigned.remove({member})
                except ValueError:
                    pass
                # find if the member is already in some team
                for i, t in enumerate(teams):
                    if member in t:
                        if team_nr is None:
                            # set the team_nr we want to append these members to
                            team_nr = i
                        elif team_nr != i:
                            # some member is in a team and this one is in another team -
                            # we have to unite those teams
                            result_team_nr = min([i, team_nr])
                            teams[result_team_nr] = teams[i] | teams[team_nr]
                            del teams[max(i, team_nr)]
                            team_nr = result_team_nr
            if team_nr is not None:
                teams[team_nr] |= team
            else:
                teams.append(team)

        for team in teams:
            # check team validity
            if len(team) > self.max_team_size or not self.checkTeamValid(team):
                raise ImpossibleDistributionError()
        # add created teams to unassigned
        self.unassigned += teams

    def _solve(self, unassigned):
        """
        generate possible distributions, that have valid teams 
        (teams with size <= max_team_size and not violating any requirements)
        Warning: the teams can still be shorter than required!
        :param unassigned:  members to be distributed
        :return: [{team1}, {team2}, ...]
        """
        if len(unassigned) == 1:
            yield [unassigned[0]]
            return
        for perm in self._solve(unassigned[1:]):
            if self.prefer:
                # sort teams, so we start adding to the best possible
                order = sorted(
                    range(len(perm)),
                    key=lambda x: sum(member.history_count[member_with] for member_with in perm[x] for member in unassigned[0]),
                    reverse=True
                )
            else:
                order = range(len(perm))
            for i in order:
                if self.canAddToTeam(unassigned[0], perm[i]):
                    yield perm[:i] + [perm[i] | unassigned[0]] + perm[i+1:]
            if self.team_count is None or len(perm) < self.team_count:
                yield perm + [unassigned[0]]

    def canAddToTeam(self, who, team):
        """
        check if member(s) can be added to team without violating max_team_size or any requirements
        :param who: members that are being added (iterable)
        :param team: (iterable)
        :return: True if can be added, False otherwise 
        """
        if len(team) + len(who) > self.max_team_size:
            return False
        for member in who:
            for member_blacklist in member.blacklist_:
                if member_blacklist in team:
                    return False
        return True

    def printTeams(self, teams):
        """
        debug function
        """
        for i, team in enumerate(teams):
            print("Team ", i, ":", sep='')
            for member in team:
                print(member.name, end=", ")
            print()
        print()